%c

%{
#include <stdio.h>
#include <string.h>
#include "builtin.h"
#include "y.tab.h"

#define LEXDEBUG 0

int find_id(char *string);

/*
"<<"[a-zA-Z_"(), ]*">>"	return T_BLIZZ_API;
*/
%}

%%

"/*"([^*]*|("*"+[^*/]))*"*"+"/"	{
    char *c;

    for (c = yytext ; *c != '\0' ; c++) {
	if (*c == '\n')
	    lineno++;
    }
}

"//".*				;

"&&"			return T_AND;
"||"			return T_OR;
"<="			return T_LEQ;
">="			return T_GEQ;
"!="			return T_NEQ;
"=="			return T_EQ;

[0-9]+ 			{ yylval.intg = atoi(yytext); return T_INTEGER; }
[0-9]*"."[0-9]+		{ yylval.fl = atof(yytext); return T_FLOAT; }

\"[^"]*\"		return T_SPELL;

[a-zA-Z_.]+		return find_id(yytext);

[ \t]			;
\n			lineno++;
.			{
			    return *yytext;

			    /*
 			     * put in a reference to find_rule so the
			     * so the compiler shuts up
			     */
			    goto find_rule;
			}


%%

struct kw {
    struct kw *next;
    char *word;
    int token;
    YYSTYPE yylval;
} *first_kw = NULL;

void
keyword_add (char *word, int token, YYSTYPE lval)
{
    struct kw *new;

    new = amalloc(sizeof(*new));
    new->word = word;
    new->token = token;
    new->yylval = lval;
    new->next = first_kw;
    first_kw = new;
}

void
lex_init (void)
{
    static struct {
        char *word;
	int token;
	enum built_in val;
    } init_keywords[] = {
	{ "absval",	T_ABSVAL },
	{ "alternate",	T_ALTERNATE },
	{ "build", 	T_BUILD },
	{ "call",	T_CALL },
	{ "case", 	T_CASE },
	{ "class", 	T_CLASS },
	{ "clip", 	T_CLIP },
	{ "default",	T_DEFAULT },
	{ "else", 	T_ELSE },
	{ "elseif", 	T_ELSEIF },
	{ "extern", 	T_EXTERN },
	{ "fail", 	T_FAIL },
	{ "free",	T_FREE },
	{ "function",	T_FUNCTION },
	{ "if", 	T_IF },
	{ "incombat", 	T_COMBAT },
	{ "mutex",	T_MUTEX },
	{ "nil",	T_NIL },
	{ "nil", 	T_NIL },
	{ "outofcombat",T_OUTOFCOMBAT },
	{ "postspell",	T_POSTSPELL },
	{ "precombat", 	T_PRECOMBAT },
	{ "print", 	T_PRINT },
	{ "printf", 	T_PRINT },
	{ "spellsent",	T_SPELLSENT },
	{ "strict", 	T_STRICT },
	{ "switch",	T_SWITCH },
	{ "var",	T_VAR },
 
	// built-in functions
	{ "boss",		T_BUILTIN_1,	BI_BOSS },
	{ "buff",		T_BUILTIN_2,	BI_BUFF },
	{ "buffInfo",		T_BUILTIN_2,	BI_BUFF_INFO },
	{ "buffTime",		T_BUILTIN_2,	BI_BUFF_TIME },
	{ "castable",		T_BUILTIN_2, 	BI_CASTABLE },
	{ "combopoints",	T_BUILTIN_1,	BI_COMBOPOINTS },
	{ "creaturetype", 	T_BUILTIN_2M,	BI_CREATURETYPE },
	{ "debuff",		T_BUILTIN_2,	BI_DEBUFF_BY_ANY },
	{ "debuffByMe",		T_BUILTIN_2,	BI_DEBUFF_BY_ME },
	{ "debuffInfo",		T_BUILTIN_2,	BI_DEBUFF_INFO },
	{ "flaggedPVP",		T_BUILTIN_1,	BI_FLAGGEDPVP },
	{ "flighttime",		T_BUILTIN_1,	BI_FLIGHTTIME },
	{ "gcd",		T_BUILTIN_2,	BI_GCD },
	{ "hasglyph",		T_BUILTIN_2, 	BI_HASGLYPH },
	{ "haspet",		T_BUILTIN_1, 	BI_HASPET },
	{ "known",		T_BUILTIN_2,	BI_KNOWSSPELL },
	{ "lastcast",		T_BUILTIN_2,	BI_LASTCAST },
	{ "mounted",		T_BUILTIN_1, 	BI_MOUNTED },
	{ "mxbuff",		T_BUILTIN_MX,	BI_MXBUFF },
	{ "mxbuffbyme",		T_BUILTIN_MX,	BI_MXBUFF_BY_ME },
	{ "mxdebuffbyme",	T_BUILTIN_MX,	BI_MXDEBUFF_BY_ME },
	{ "mxneed",		T_BUILTIN_MX2,	BI_MXNEED },
	{ "happypet",		T_BUILTIN_1,	BI_PETHAPPINESS },
	{ "petbuff", 		T_BUILTIN_2,	BI_PETBUFF },
	{ "petexists",		T_BUILTIN_1, 	BI_PETEXISTS },
	{ "pethealth", 		T_BUILTIN_1,	BI_PETHEALTH },
	{ "power", 		T_BUILTIN_1,	BI_POWER },
	{ "threat",		T_BUILTIN_1,	BI_THREAT },
	{ "totemfire",		T_BUILTIN_1,	BI_TOTEMFIRE },
	{ "rangedcombat",	T_BUILTIN_2M, 	BI_RANGED },
	{ "weaponenhancement",	T_BUILTIN_2,	BI_WEAPONENHANCE },
    	{ "playerhealth",	T_BUILTIN_1,	BI_PLAYERHEALTH, },
    	{ "targethealth",	T_BUILTIN_1,	BI_TARGETHEALTH, },
    	{ "timetodie",		T_BUILTIN_1,	BI_TIMETODIE, },
    };
    int i;

    for (i = 0 ; i < sizeof(init_keywords) / sizeof(*init_keywords) ; i++) {
	YYSTYPE lval;

	lval.bi = init_keywords[i].val;
	keyword_add(init_keywords[i].word, init_keywords[i].token, lval);
    }
}

void
lex_register_variable (char *name, int token)
{
    YYSTYPE lval;

    lval.str = astrdup(name);
    keyword_add(lval.str, token, lval);
}

void
lex_register_mutex (char *name, int mindex)
{
    YYSTYPE lval;

    lval.intg = mindex;
    keyword_add(astrdup(name), T_MUTEXVAR, lval);
}

int
find_id (char *string)
{
    struct kw *cur;

    // linear search: efficient enough for these few entries
    for (cur = first_kw ; cur != NULL ; cur = cur->next) {
	if (!strcasecmp(string, cur->word)) {
	   yylval = cur->yylval;
#if LEXDEBUG
	   fflush(stdout);
	   fprintf(stderr, "<find_id(%s) returns %d>\n", string, cur->token);
#endif	   
	   return cur->token;
	}
    }
#if LEXDEBUG
    fflush(stdout);
    fprintf(stderr, "<find_id(%s) returns T_ID>\n", string);
#endif	   
    return T_ID;
}

// make the compiler shut up about unused references by referencing them all
void
unused_func (void)
{
    void *vs = NULL;
    yy_size_t s = 0;
    char c = 2;
    
    yy_flex_realloc(vs, s);
    yyunput(c, &c);
    yy_flex_strlen("foo");
}

